from . import Lifter, register

class FixesPostProcessor(Lifter):
    def postprocess(self):
        self.irsb.statements = [x for x in self.irsb.statements if x.tag != 'Ist_NoOp']

        funcname = "_post_process_%s" % self.irsb.arch.name
        if hasattr(self, funcname):
            getattr(self, funcname)()

    def _post_process_ARM(self):
        # Jumpkind
        if self.irsb.jumpkind == "Ijk_Boring":
            # If PC is moved to LR, then this should be an Ijk_Call
            #
            # Example:
            # MOV LR, PC
            # MOV PC, R8

            # Note that the value of PC is directly used in IRStatements, i.e
            # instead of having:
            #   t0 = GET:I32(pc)
            #   PUT(lr) = t0
            # we have:
            #   PUT(lr) = 0x10400
            # The only case (that I've seen so far) where a temporary variable
            # is assigned to LR is:
            #   t2 = ITE(cond, t0, t1)
            #   PUT(lr) = t2

            pc_holders = {}
            lr_store_pc = False
            inst_ctr = 0
            next_irsb_addr = self.irsb.statements[0].addr + self.irsb.size
            for stt in self.irsb.statements:
                if type(stt) == stmt.Put:
                    # LR is modified just before the last instruction of the
                    # block...
                    if stt.offset == self.irsb.arch.registers['lr'][0] \
                       and inst_ctr == self.irsb.instructions - 1:
                        # ... by a constant, so test whether it is the address
                        # of the next IRSB
                        if type(stt.data) == expr.Const:
                            if stt.data.con.value == next_irsb_addr:
                                lr_store_pc = True
                        # ... by a temporary variable, so test whether it holds
                        # the address of the next IRSB
                        elif type(stt.data) == expr.RdTmp:
                            if next_irsb_addr == pc_holders.get(stt.data.tmp):
                                lr_store_pc = True
                        break
                    else:
                        reg_name = self.irsb.arch.translate_register_name(stt.offset)
                        if type(stt.data) == expr.Const:
                            pc_holders[reg_name] = stt.data.con.value
                        elif type(stt.data) == expr.RdTmp and pc_holders.get(stt.data.tmp) is not None:
                            pc_holders[reg_name] = pc_holders[stt.data.tmp]
                        elif type(stt.data) == expr.Get and pc_holders.get(stt.data.offset) is not None:
                            pc_holders[reg_name] = pc_holders[stt.data.offset]
                elif type(stt) == stmt.WrTmp:
                    # the PC value may propagate through the block, and since
                    # LR is modified at the end of the block, the PC value have
                    # to be incremented in order to match the address of the
                    # next IRSB. So the only propagation ways that can lead to
                    # a function call are:
                    #   - Iop_Add* operations (even "sub r0, #-4" is compiled
                    #   as "add r0, #4")
                    #   - Iop_And*, Iop_Or*, Iop_Xor*, Iop_Sh*, Iop_Not* (there
                    #   may be some tricky and twisted ways to increment PC)
                    if type(stt.data) in (expr.Unop, expr.Binop, expr.Triop, expr.Qop):
                        if all(type(a) == expr.Const
                               or (type(a) == expr.RdTmp and pc_holders.get(a.tmp) is not None)
                                    for a in stt.data.args):
                            op = stt.data.op
                            vals = [a.con.value if type(a) == expr.Const else pc_holders[a.tmp] \
                                    for a in stt.data.args]
                            if 'Iop_Add' in op:
                                pc_holders[stt.tmp] = sum(vals)
                            elif 'Iop_And' in op:
                                pc_holders[stt.tmp] = reduce(lambda a, b: a & b, vals)
                            elif 'Iop_Or' in op:
                                pc_holders[stt.tmp] = reduce(lambda a, b: a | b, vals)
                            elif 'Iop_Xor' in op:
                                pc_holders[stt.tmp] = reduce(lambda a, b: a ^ b, vals)
                            elif 'Iop_Shl' in op:
                                pc_holders[stt.tmp] = vals[0] << vals[1]
                            elif any(o in op for o in ('Iop_Shr', 'Iop_Sar')):
                                pc_holders[stt.tmp] = vals[0] >> vals[1]
                    elif type(stt.data) == expr.Get:
                        reg_name = self.irsb.arch.translate_register_name(stt.data.offset)
                        if pc_holders.get(reg_name) is not None:
                            pc_holders[stt.tmp] = pc_holders[reg_name]
                    elif type(stt.data) == expr.ITE:
                        for d in (stt.data.iffalse, stt.data.iftrue):
                            if type(d) == expr.Const:
                                pc_holders[stt.tmp] = d.con.value
                            elif type(d) == expr.RdTmp and pc_holders.get(d.tmp) is not None:
                                pc_holders[stt.tmp] = pc_holders[d.tmp]
                    elif type(stt.data) == expr.RdTmp and pc_holders.get(stt.data.tmp) is not None:
                        pc_holders[stt.tmp] = pc_holders[stt.data.tmp]
                    elif type(stt.data) == expr.Const:
                        pc_holders[stt.tmp] = stt.data.con.value

                elif type(stt) == stmt.IMark:
                    inst_ctr += 1

            if lr_store_pc:
                self.irsb.jumpkind = 'Ijk_Call'

    _post_process_ARMEL = _post_process_ARM
    _post_process_ARMHF = _post_process_ARM

    def _post_process_MIPS32(self):
        # Handle unconditional branches
        # `beq $zero, $zero, xxxx`
        # It is translated to
        #
        # 15 | ------ IMark(0x401684, 4, 0) ------
        # 16 | t0 = CmpEQ32(0x00000000, 0x00000000)
        # 17 | PUT(128) = 0x00401688
        # 18 | ------ IMark(0x401688, 4, 0) ------
        # 19 | if (t0) goto {Ijk_Boring} 0x401684
        # 20 | PUT(128) = 0x0040168c
        # 21 | t4 = GET:I32(128)
        # NEXT: PUT(128) = t4; Ijk_Boring
        #

        stts = self.irsb.statements
        tmp_exit = None
        exit_stt_idx = None
        dst = None

        for i, stt in reversed(list(enumerate(stts))):
            if tmp_exit is None:
                # Looking for the Exit statement
                if isinstance(stt, stmt.Exit) and \
                        isinstance(stt.guard, expr.RdTmp):
                    tmp_exit = stt.guard.tmp
                    dst = stt.dst
                    exit_stt_idx = i
            else:
                # Looking for the WrTmp statement
                if isinstance(stt, stmt.WrTmp) and \
                                stt.tmp == tmp_exit:
                    if isinstance(stt.data, expr.Binop) and \
                                    stt.data.op == 'Iop_CmpEQ32' and \
                            isinstance(stt.data.child_expressions[0], expr.Const) and \
                            isinstance(stt.data.child_expressions[1], expr.Const) and \
                                    stt.data.child_expressions[0].con.value == stt.data.child_expressions[
                                1].con.value:

                        # Create a new IRConst
                        irconst = expr.Const.__new__(expr.Const)    # XXX: does this work???
                        irconst.con = dst
                        irconst.tag = 'Iex_Const'

                        self.irsb.statements = self.irsb.statements[: exit_stt_idx] + self.irsb.statements[exit_stt_idx + 1:]
                        # Replace the default exit!
                        self.irsb.next = irconst

                    else:
                        break

register(FixesPostProcessor)

from .. import stmt, expr



import re
import collections

from . import pvc, ffi

class VEXObject(object):
    """
    The base class for Vex types.
    """

    __slots__ = [ ]

class IRCallee(VEXObject):
    """
    Describes a helper function to call.
    """

    __slots__ = ['regparms', 'name', 'mcx_mask', 'addr']

    def __init__(self, regparms, name, addr, mcx_mask):
        VEXObject.__init__(self)
        self.regparms = regparms
        self.name = name
        self.addr = addr
        self.mcx_mask = mcx_mask

    def __str__(self):
        return self.name

    @staticmethod
    def _from_c(c_callee):
        return IRCallee(c_callee.regparms,
                        ffi.string(c_callee.name),
                        int(ffi.cast("unsigned long long", c_callee.addr)),
                        c_callee.mcx_mask)

    @staticmethod
    def _to_c(callee):
        c_callee = pvc.mkIRCallee(callee.regparms,
                                  callee.name,
                                  ffi.cast("void *", callee.addr))
        c_callee.mcx_mask = callee.mcx_mask
        return c_callee


class IRRegArray(VEXObject):
    """
    A section of the guest state that we want te be able to index at run time, so as to be able to describe indexed or
    rotating register files on the guest.

    :ivar int base:     The offset into the state that this array starts
    :ivar str elemTy:   The types of the elements in this array, as VEX enum strings
    :ivar int nElems:   The number of elements in this array
    """

    __slots__ = ['base', 'elemTy', 'nElems']

    def __init__(self, base, elemTy, nElems):
        VEXObject.__init__(self)
        self.base = base
        self.elemTy = elemTy
        self.nElems = nElems

    def __str__(self):
        return "%s:%sx%d" % (self.base, self.elemTy[4:], self.nElems)

    @staticmethod
    def _from_c(c_arr):
        return IRRegArray(c_arr.base,
                          ints_to_enums[c_arr.elemTy],
                          c_arr.nElems)

    @staticmethod
    def _to_c(arr):
        return pvc.mkIRRegArray(arr.base,
                                get_int_from_enum(arr.elemTy),
                                arr.nElems)

ints_to_enums = { }
enums_to_ints = { }
irop_enums_to_ints = { }
will_be_overwritten = ['Ircr_GT', 'Ircr_LT']

def get_enum_from_int(i):
    return ints_to_enums[i]

def get_int_from_enum(e):
    return enums_to_ints[e]

def _add_enum(s, i=None): # TODO get rid of this
    if i is None:
        while _add_enum.counter in ints_to_enums:
            _add_enum.counter += 1
        i = _add_enum.counter
        _add_enum.counter += 1 # Update for the next iteration
    if i in ints_to_enums:
        if ints_to_enums[i] not in will_be_overwritten:
            raise ValueError('Enum with intkey %d already present' % i)
    enums_to_ints[s] = i
    ints_to_enums[i] = s
    if s.startswith('Iop_'):
        irop_enums_to_ints[s] = i
_add_enum.counter = 0

for attr in dir(pvc):
    if hasattr(pvc, attr) and isinstance(getattr(pvc, attr), int):
        _add_enum(attr, getattr(pvc, attr))

def vex_endness_from_string(endness_str):
    return getattr(pvc, endness_str)

def default_vex_archinfo():
    return {
        'hwcaps': 0,
        'endness': vex_endness_from_string('VexEndnessLE'),
        'ppc_icache_line_szB': 0,
        'ppc_dcbz_szB': 0,
        'ppc_dcbzl_szB': 0,
        'arm64_dMinLine_lg2_szB': 0,
        'arm64_iMinLine_lg2_szB': 0,
        'hwcache_info': {
            'num_levels': 0,
            'num_caches': 0,
            'caches': None,
            'icaches_maintain_coherence': True,
        },
        'x86_cr0': 0xffffffff
    }



import threading
import logging
l = logging.getLogger('pyvex.lift.libvex')
l.setLevel(20) # Shut up

from . import Lifter, register
from .. import pvc, ffi

_libvex_lock = threading.Lock()

SUPPORTED = {'X86', 'AMD64', 'MIPS32', 'MIPS64', 'ARM', 'ARMEL', 'ARMHF', 'AARCH64', 'PPC32', 'PPC64'}

class LibVEXLifter(Lifter):
    REQUIRE_DATA_C = True

    def lift(self):
        if self.irsb.arch.name not in SUPPORTED:
            return False

        if self.traceflags != 0 and l.getEffectiveLevel() > 20:
            l.setLevel(20)

        try:
            _libvex_lock.acquire()
            self.irsb.arch.vex_archinfo['hwcache_info']['caches'] = ffi.NULL

            pvc.log_level = l.getEffectiveLevel()
            vex_arch = getattr(pvc, self.irsb.arch.vex_arch)

            if self.max_inst is None: self.max_inst = 99
            if self.max_bytes is None: self.max_bytes = 5000
            c_irsb = pvc.vex_lift(vex_arch, self.irsb.arch.vex_archinfo, self.data + self.bytes_offset, self.irsb._addr, self.max_inst, self.max_bytes, self.opt_level, self.traceflags, self.allow_lookback)

            log_str = str(ffi.buffer(pvc.msg_buffer, pvc.msg_current_size)) if pvc.msg_buffer != ffi.NULL else None

            if c_irsb == ffi.NULL:
                self._error = "libvex: unkown error" if log_str is None else log_str
                return False
            else:
                if log_str is not None:
                    l.info(log_str)

            self.irsb._from_c(c_irsb)
        finally:
            _libvex_lock.release()
            # We must use a pickle value, CData objects are not pickeable so not ffi.NULL
            self.irsb.arch.vex_archinfo['hwcache_info']['caches'] = None

        return True

register(LibVEXLifter)



"""
PyVEX provides an interface that translates binary code into the VEX intermediate represenation (IR).
For an introduction to VEX, take a look here: https://docs.angr.io/docs/ir.html
"""

import os
import sys
import cffi
import pkg_resources

from .vex_ffi import ffi_str as _ffi_str
ffi = cffi.FFI()

import logging
logging.getLogger("pyvex").addHandler(logging.NullHandler())

def _find_c_lib():
    # Load the c library for calling into VEX
    if sys.platform in ('win32', 'cygwin'):
        library_file = 'pyvex.dll'
    elif sys.platform == 'darwin':
        library_file = "libpyvex.dylib"
    else:
        library_file = "libpyvex.so"

    pyvex_path = pkg_resources.resource_filename(__name__, os.path.join('lib', library_file))

    ffi.cdef(_ffi_str)
    # RTLD_GLOBAL used for sim_unicorn.so
    lib = ffi.dlopen(pyvex_path)
    lib.vex_init()
    # this looks up all the definitions (wtf)
    dir(lib)
    return lib

pvc = _find_c_lib()

# pylint: disable=wildcard-import
from .enums import *
from .block import *
from . import stmt, expr, const
from .expr import get_op_retty
from .const import tag_to_const_class

# aliases....
IRStmt = stmt
IRExpr = expr
IRConst = const



import re

from . import VEXObject, ffi, pvc
from .enums import get_enum_from_int
from .errors import PyVEXError

# IRConst hierarchy
class IRConst(VEXObject):

    __slots__ = ['value']

    type = None
    tag = None
    c_constructor = None

    def __init__(self):
        VEXObject.__init__(self)

    def pp(self):
        print self.__str__()

    @property
    def size(self):
        return get_type_size(self.type)

    @staticmethod
    def _from_c(c_const):
        if c_const[0] == ffi.NULL:
            return None

        tag = get_enum_from_int(c_const.tag)

        try:
            return tag_to_const_class(tag)._from_c(c_const)
        except KeyError:
            raise PyVEXError('Unknown/unsupported IRConstTag %s\n' % tag)
    _translate = _from_c

    @classmethod
    def _to_c(cls, const):
        # libvex throws an exception when constructing a U1 with a value other than 0 or 1
        if const.tag == 'Ico_U1' and not const.value in (0, 1):
            raise PyVEXError('Invalid U1 value: %d' % const.value)

        try:
            return cls.c_constructor(const.value)
        except KeyError:
            raise PyVEXError('Unknown/unsupported IRConstTag %s]n' % const.tag)

class U1(IRConst):
    type = 'Ity_I1'
    tag = 'Ico_U1'
    op_format = '1'
    c_constructor = pvc.IRConst_U1

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "%d" % self.value

    @staticmethod
    def _from_c(c_const):
        return U1(c_const.Ico.U1)

class U8(IRConst):
    type = 'Ity_I8'
    tag = 'Ico_U8'
    op_format = '8'
    c_constructor = pvc.IRConst_U8

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "0x%02x" % self.value

    @staticmethod
    def _from_c(c_const):
        return U8(c_const.Ico.U8)

class U16(IRConst):
    type = 'Ity_I16'
    tag = 'Ico_U16'
    op_format = '16'
    c_constructor = pvc.IRConst_U16

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "0x%04x" % self.value

    @staticmethod
    def _from_c(c_const):
        return U16(c_const.Ico.U16)

class U32(IRConst):
    type = 'Ity_I32'
    tag = 'Ico_U32'
    op_format = '32'
    c_constructor = pvc.IRConst_U32

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "0x%08x" % self.value

    @staticmethod
    def _from_c(c_const):
        return U32(c_const.Ico.U32)

class U64(IRConst):
    type = 'Ity_I64'
    tag = 'Ico_U64'
    op_format = '64'
    c_constructor = pvc.IRConst_U64

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "0x%016x" % self.value

    @staticmethod
    def _from_c(c_const):
        return U64(c_const.Ico.U64)

# Integer Type Imagination
class_cache = { 1 : U1, 8 : U8, 16 : U16, 32 : U32, 64 : U64 }

def vex_int_class(size):
    try:
        return class_cache[size]
    except KeyError:
        class VexInt(IRConst):
            type = 'Ity_I%d' % size
            tag = 'Ico_U%d' % size
            op_format = str(size)

            def __init__(self, value):
                IRConst.__init__(self)
                self.value = value

            def __str__(self):
                return '(0x%x :: %s)' % (self.value, self.type)
        VexInt.__name__ = 'U%d' % size
        class_cache[size] = VexInt
        return VexInt

class F32(IRConst):
    type = 'Ity_F32'
    tag = 'Ico_F32'
    op_format = 'F32'
    c_constructor = pvc.IRConst_F32

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "%f" % self.value

    @staticmethod
    def _from_c(c_const):
        return F32(c_const.Ico.F32)

class F32i(IRConst):
    type = 'Ity_F32'
    tag = 'Ico_F32i'
    op_format = 'F32'
    c_constructor = pvc.IRConst_F32i

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "%f" % self.value

    @staticmethod
    def _from_c(c_const):
        return F32i(c_const.Ico.F32)

class F64(IRConst):
    type = 'Ity_F64'
    tag = 'Ico_F64'
    op_format = 'F64'
    c_constructor = pvc.IRConst_F64

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "%f" % self.value

    @staticmethod
    def _from_c(c_const):
        return F64(c_const.Ico.F64)

class F64i(IRConst):
    type = 'Ity_F64'
    tag = 'Ico_F64i'
    op_format = 'F64'
    c_constructor = pvc.IRConst_F64i

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "%f" % self.value

    @staticmethod
    def _from_c(c_const):
        return F64i(c_const.Ico.F64)

class V128(IRConst):
    type = 'Ity_V128'
    tag = 'Ico_V128'
    op_format = 'V128'
    c_constructor = pvc.IRConst_V128

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "%x" % self.value

    @staticmethod
    def _from_c(c_const):
        return V128(c_const.Ico.V128)

class V256(IRConst):
    type = 'Ity_V256'
    tag = 'Ico_V256'
    op_format = 'V256'
    c_constructor = pvc.IRConst_V256

    def __init__(self, value):
        IRConst.__init__(self)
        self.value = value

    def __str__(self):
        return "%x" % self.value

    @staticmethod
    def _from_c(c_const):
        return V256(c_const.Ico.V256)

predefined_types = [ U1, U8, U16, U32, U64, F32, F32i, F64, F64i, V128, V256 ]
predefined_types_map = { c.type : c for c in predefined_types }
predefined_classes_map = { c.tag : c for c in predefined_types }

def is_int_ty(ty):
    m = re.match(r'Ity_I\d+', ty)
    return m is not None

def is_int_tag(tag):
    m = re.match(r'Ico_U\d+', tag)
    return m is not None

def get_tag_size(tag):
    m = re.match(r'Ico_[UFV](?P<size>\d+)i?', tag)
    if m is None:
        raise ValueError('Tag %s does not have size' % ty)
    return int(m.group('size'))

def get_type_size(ty):
    m = re.match(r'Ity_[IFDV](?P<size>\d+)', ty)
    if m is None:
        raise ValueError('Type %s does not have size' % ty)
    return int(m.group('size'))

def ty_to_const_class(ty):
    try:
        return predefined_types_map[ty]
    except KeyError:
        if is_int_ty(ty):
            size = get_type_size(ty)
            return vex_int_class(size)
        else:
            raise ValueError('Type %s does not exist' % ty)

def tag_to_const_class(tag):
    try:
        return predefined_classes_map[tag]
    except KeyError:
        if is_int_tag(tag):
            size = get_tag_size(tag)
            return vex_int_class(size)
        else:
            raise ValueError('Tag %s does not exist' % tag)



import struct
import archinfo
import pyvex
from functools import wraps
from pyvex.expr import *
from pyvex.lift import Lifter, register
from .vex_helper import *
from syntax_wrapper import VexValue
import logging
import bitstring


def is_empty(bitstrm):
    try:
        bitstrm.peek(1)
        return False
    except bitstring.ReadError:
        return True


class ParseError(Exception):
    pass


class GymratLifter(Lifter):
    """
    This is a base class for lifters that use Gymrat.
    For most architectures, all you need to do is subclass this, and set the property "instructions"
    to be a list of classes that define each instruction.
    By default, a lifter will decode instructions by attempting to instantiate every class until one works.
    This will use an IRSBCustomizer, which will, if it succeeds, add the appropriate VEX instructions to a pyvex IRSB.
    pyvex, when lifting a block of code for this architecture, will call the method "lift", which will produce the IRSB
    of the lifted code.
    """
    def __init__(self, irsb, data, max_inst, max_bytes, bytes_offset, opt_level=None,
                    traceflags=None, allow_lookback=None):
        super(GymratLifter, self).__init__(irsb, data, max_inst, max_bytes,
                bytes_offset, opt_level, traceflags, allow_lookback)
        self.logger = logging.getLogger('lifter')
        self.logger.setLevel(logging.DEBUG)
        if 'CData' in str(type(data)):
            thedata = "".join([chr(data[x]) for x in range(max_bytes)])
        else:
            thedata = data
        self.bitstrm = bitstring.ConstBitStream(bytes=thedata)

    def lift(self, disassemble=False, dump_irsb=False):
        try:
            count = 0
            disas = []
            if not self.max_inst:
                self.max_inst = 1000000
            self.irsb.jumpkind = JumpKind.Invalid
            irsb_c = IRSBCustomizer(self.irsb, self.irsb.arch)
            addr = self.irsb._addr
            pos = self.bitstrm.bitpos
            while (count < self.max_inst
                    and (self.bitstrm.bitpos + 7) / 8 < self.max_bytes
                    and self.irsb.jumpkind == JumpKind.Invalid
                    and not is_empty(self.bitstrm)):
                # Try every instruction until one works
                for possible_instr in self.instrs:
                    try:
                        instr = possible_instr(irsb_c, self.bitstrm, addr)
                        break
                    except ParseError:
                        pass #self.logger.exception(repr(possible_instr))
                    except Exception, e:
                        self.logger.debug(e.message)
                else:
                    errorstr = 'Unknown instruction at bit position %d' % self.bitstrm.bitpos
                    self.logger.critical(errorstr)
                    self.logger.critical("Address: %#08x" % addr)
                    raise Exception(errorstr)
                if disassemble:
                    disas.append(instr.disassemble())
                else:
                    instr()
                # WARNING: We assume 8 bit bytes here.
                addr += (self.bitstrm.bitpos - pos) / 8
                pos = self.bitstrm.bitpos
                count += 1
            if dump_irsb:
                self.irsb.pp()
            if disassemble:
                return disas
            return self.irsb
        except Exception, e:
            self.errors = e.message
            self.logger.exception("Error lifting block:")
            raise e

    def pp_disas(self, addr, name, args):
        args_str = ",".join(str(a) for a in args)
        return "%0#08x:\t%s %s" % (addr, name, args_str)

    def error(self):
        return self.errors

    def disassemble(self):
        return self.lift(disassemble=True)

if __name__ == '__main__':
    pass



import functools

from .vex_helper import IRSBCustomizer, Type, JumpKind
import pyvex
from pyvex.expr import IRExpr, Const, RdTmp, Unop, Binop, Triop, Qop, Load, CCall, Get


def checkparams(*fn, **options):
    rhstype = options.pop('rhstype', None)
    if options:
        raise TypeError("unsupported keyword arguments: %s" %
                        ",".join(options.keys()))

    if fn:
        def wrapper(self, *args, **kwargs):
            irsb_cs = {a.irsb_c for a in list(args) + [self] if isinstance(a, VexValue)}  # pylint: disable=no-member
            assert len(irsb_cs) == 1, 'All VexValues must belong to the same irsb_c'
            args = list(args)
            for arg in args:
                if isinstance(arg, int):
                    thetype = rhstype if rhstype else self.ty
                    args[args.index(arg)] = VexValue.Constant(self.irsb_c, arg, thetype)
            args = tuple(args)
            return fn[0](self, *args, **kwargs)
        return functools.update_wrapper(wrapper, fn[0])
    else:
        def decorator(fn):
            def wrapper(self, *args, **kwargs):
                irsb_cs = {a.irsb_c for a in list(args) + [self] if
                           isinstance(a, VexValue)}  # pylint: disable=no-member
                assert len(irsb_cs) == 1, 'All VexValues must belong to the same irsb_c'
                args = list(args)
                for arg in args:
                    if isinstance(arg, int):
                        thetype = rhstype if rhstype else self.ty
                        args[args.index(arg)] = VexValue.Constant(self.irsb_c, arg, thetype)
                args = tuple(args)
                return fn(self, *args, **kwargs)
            return functools.update_wrapper(wrapper, fn)
        return decorator


def vvifyresults(f):
    @functools.wraps(f)
    def decor(self, *args, **kwargs):
        returned = f(self, *args, **kwargs)
        assert isinstance(returned, RdTmp) or isinstance(returned, Const)
        return VexValue(self.irsb_c, returned)
    return decor


class VexValue:
    def __init__(self, irsb_c, rdt):
        self.irsb_c = irsb_c
        self.ty = self.irsb_c.get_type(rdt)
        self.rdt = rdt
        self.width = pyvex.get_type_size(self.ty)
        self.is_signed = False

    @property
    def value(self):
        if isinstance(self.rdt, Const):
            return self.rdt.con.value
        else:
            raise ValueError("Non-constant VexValue has no value property")

    def toSigned(self):
        self.is_signed = True

    def toUnsigned(self):
        self.is_signed = False

    @vvifyresults
    def widen_unsigned(self, ty):
        return self.irsb_c.op_widen_int_unsigned(self.rdt, ty)

    @vvifyresults
    def cast_to(self, ty, signed=False, high=False):
        return self.irsb_c.cast_to(self.rdt, ty, signed=signed, high=high)

    @vvifyresults
    def widen_signed(self, ty):
        return self.irsb_c.op_widen_int_signed(self.rdt, ty)

    @vvifyresults
    def narrow_high(self, ty):
        return self.irsb_c.op_narrow_int(self.rdt, ty, high_half=True)

    @vvifyresults
    def narrow_low(self, ty):
        return self.irsb_c.op_narrow_int(self.rdt, ty, high_half=False)

    # TODO at some point extend this to Vex nonconstants
    def __getitem__(self, idx):
        getb = lambda i: VexValue(self.irsb_c, self.irsb_c.get_bit(self.rdt, i))
        makeconstant = lambda x: VexValue.Constant(self.irsb_c, x, Type.int_8).rdt
        if not isinstance(idx, slice):
            actualindex = slice(idx).indices(self.width)[1]
            return getb(makeconstant(actualindex))
        else:
            return [getb(makeconstant(i)) for i in xrange(*idx.indices(self.width))]

    @checkparams
    @vvifyresults
    def set_bit(self, idx, bval):
        typedidx = idx.cast_to(Type.int_8)
        return self.irsb_c.set_bit(self.rdt, idx.rdt, bval.rdt)

    @checkparams
    @vvifyresults
    def set_bits(self, idxsandvals):
        return self.irsb_c.set_bits(self.rdt, [(i.cast_to(Type.int_8).rdt, b.rdt) for i, b in idxsandvals])

    @checkparams
    @vvifyresults
    def ite(self, iftrue, iffalse):
        onebitcond = self.cast_to(Type.int_1)
        return self.irsb_c.ite(onebitcond.rdt, iftrue.rdt, iffalse.rdt)

    @checkparams
    @vvifyresults
    def __add__(self, right):
        return self.irsb_c.op_add(self.rdt, right.rdt)

    @checkparams
    @vvifyresults
    def __sub__(self, right): # I'm not sure if we want rsub
        return self.irsb_c.op_sub(self.rdt, right.rdt)

    @checkparams
    @vvifyresults
    def __div__(self, right):
        return self.irsb_c.op_div(self.rdt, right.rdt, self.is_signed)

    @checkparams
    @vvifyresults
    def __floordiv__(self, right): # Note: nonprimitive
        return self / right

    @checkparams
    @vvifyresults
    def __truediv__(self, right): # Note: nonprimitive
        return self / right

    @checkparams
    @vvifyresults
    def __and__(self, right):
        return self.irsb_c.op_and(self.rdt, right.rdt)

    @checkparams
    @vvifyresults
    def __eq__(self, right):
        return self.irsb_c.op_cmp_eq(self.rdt, right.rdt)

    @checkparams
    @vvifyresults
    def __ne__(self, other):
        return self.irsb_c.op_cmp_ne(self.rdt, other.rdt)

    @checkparams
    @vvifyresults
    def __invert__(self):
        return self.irsb_c.op_not(self.rdt)

    @checkparams
    @vvifyresults
    def __le__(self, right):
        if self.is_signed:
            return self.irsb_c.op_cmp_sle(self.rdt, right.rdt)
        else:
            return self.irsb_c.op_cmp_ule(self.rdt, right.rdt)

    @checkparams
    @vvifyresults
    def __gt__(self, other):
        if self.is_signed:
            return self.irsb_c.op_cmp_sgt(self.rdt, other.rdt)
        else:
            return self.irsb_c.op_cmp_ugt(self.rdt, other.rdt)

    @checkparams
    @vvifyresults
    def __ge__(self, right):
        if self.is_signed:
            return self.irsb_c.op_cmp_sge(self.rdt, right.rdt)
        else:
            return self.irsb_c.op_cmp_uge(self.rdt, right.rdt)

    @checkparams(rhstype=Type.int_8)
    @vvifyresults
    def __lshift__(self, right):
        return self.irsb_c.op_shl(self.rdt, right.rdt)

    @checkparams
    @vvifyresults
    def __lt__(self, right):
        if self.is_signed:
            return self.irsb_c.op_cmp_slt(self.rdt, right.rdt)
        else:
            return self.irsb_c.op_cmp_ult(self.rdt, right.rdt)

    @checkparams
    @vvifyresults
    def __mod__(self, right): # Note: nonprimitive
        return self.rdt - (self.rdt / right.rdt) * right.rdt

    @checkparams
    @vvifyresults
    def __mul__(self, right):
        return self.irsb_c.op_mull(self.rdt, right.rdt, signed=self.is_signed)

    @checkparams
    @vvifyresults
    def __neg__(self): # Note: nonprimitive
        if self.is_unsigned:
            raise Exception('Number is unsigned, cannot change sign!')
        else:
            return self.rdt * -1

    @checkparams
    @vvifyresults
    def __or__(self, right):
        return self.irsb_c.op_or(self.rdt, right.rdt)

    @checkparams
    @vvifyresults
    def __pos__(self):
        return self

    @checkparams
    @vvifyresults
    def __rshift__(self, right):
        return self.irsb_c.op_shr(self.rdt, right.rdt)

    @checkparams
    @vvifyresults
    def __xor__ (self, right):
        return self.irsb_c.op_xor(self.rdt, right.rdt)

    @classmethod
    def Constant(cls, irsb_c, val, ty):
        """
        Creates a constant as a VexValue
        :param irsb_c: The IRSBCustomizer to use
        :param val: The value, as an integer
        :param ty: The type of the resulting VexValue
        :return: a VexValue
        """
        assert not (isinstance(val, VexValue) or isinstance(val, IRExpr))
        rdt = irsb_c.mkconst(val, ty)
        return cls(irsb_c, rdt)




import re
import logging

from . import VEXObject
from .enums import IRCallee, IRRegArray, get_int_from_enum, get_enum_from_int
from .const import get_type_size

l = logging.getLogger("pyvex.expr")

class IRExpr(VEXObject):
    """
    IR expressions in VEX represent operations without side effects.
    """

    tag = None

    def __init__(self):
        VEXObject.__init__(self)

    def pp(self):
        print self.__str__()

    @property
    def child_expressions(self):
        """
        A list of all of the expressions that this expression ends up evaluating.
        """
        expressions = [ ]
        for k in self.__slots__:
            v = getattr(self, k)
            if isinstance(v, IRExpr):
                expressions.append(v)
                expressions.extend(v.child_expressions)
        return expressions

    @property
    def constants(self):
        """
        A list of all of the constants that this expression ends up using.
        """
        constants = [ ]
        for k in self.__slots__:
            v = getattr(self, k)
            if isinstance(v, IRExpr):
                constants.extend(v.constants)
            elif isinstance(v, IRConst):
                constants.append(v)
        return constants

    def result_size(self, tyenv):
        return get_type_size(self.result_type(tyenv))

    def result_type(self, tyenv):
        raise NotImplementedError

    @staticmethod
    def _from_c(c_expr):
        if c_expr == ffi.NULL or c_expr[0] == ffi.NULL:
            return None

        tag = get_enum_from_int(c_expr.tag)

        try:
            return tag_to_expr_class(tag)._from_c(c_expr)
        except KeyError:
            raise PyVEXError('Unknown/unsupported IRExprTag %s\n' % tag)
    _translate = _from_c

    @staticmethod
    def _to_c(expr):
        try:
            return tag_to_expr_class(expr.tag)._to_c(expr)
        except KeyError:
            raise PyVEXError('Unknown/unsupported IRExprTag %s\n' % expr.tag)

    def typecheck(self, tyenv):
        return self.result_type(tyenv)


class Binder(IRExpr):
    """
    Used only in pattern matching within Vex. Should not be seen outside of Vex.
    """

    __slots__ = ['binder']

    tag = 'Iex_Binder'

    def __init__(self, binder):
        IRExpr.__init__(self)
        self.binder = binder

    def __str__(self):
        return "Binder"

    @staticmethod
    def _from_c(c_expr):
        return Binder(c_expr.iex.Binder.binder)

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Binder(expr.binder)

    def result_type(self, tyenv):
        return 'Ity_INVALID'

class VECRET(IRExpr):

    tag = 'Iex_VECRET'

    def __init__(self):
        IRExpr.__init__(self)

    def __str__(self):
        return "VECRET"

    @staticmethod
    def _from_c(c_expr):
        return VECRET()

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_VECRET()

    def result_type(self, tyenv):
        return 'Ity_INVALID'


class GSPTR(IRExpr):

    tag = 'Iex_GSPTR'

    def __init__(self):
        IRExpr.__init__(self)

    def __str__(self):
        return "GSPTR"

    @staticmethod
    def _from_c(c_expr):
        return GSPTR()

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_GSPTR()

    def result_type(self, tyenv):
        return 'Ity_INVALID'


class GetI(IRExpr):
    """
    Read a guest register at a non-fixed offset in the guest state.
    """

    __slots__ = ['descr', 'ix', 'bias']

    tag = 'Iex_GetI'

    def __init__(self, descr, ix, bias):
        IRExpr.__init__(self)
        self.descr = descr
        self.ix = ix
        self.bias = bias

    @property
    def description(self):
        return self.descr

    @property
    def index(self):
        return self.ix

    def __str__(self):
        return "GetI(%s)[%s,%s]" % (self.descr, self.ix, self.bias)

    @staticmethod
    def _from_c(c_expr):
        descr = IRRegArray._from_c(c_expr.Iex.GetI.descr)
        ix = IRExpr._from_c(c_expr.Iex.GetI.ix)
        bias = c_expr.Iex.GetI.bias
        return GetI(descr, ix, bias)

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_GetI(IRRegArray._to_c(expr.descr),
                               IRExpr._to_c(expr.ix),
                               expr.bias)

    def result_type(self, tyenv):
        return self.descr.elemTy


class RdTmp(IRExpr):
    """
    Read the value held by a temporary.
    """

    __slots__ = ['tmp']

    tag = 'Iex_RdTmp'

    def __init__(self, tmp):
        IRExpr.__init__(self)
        self.tmp = tmp

    def __str__(self):
        return "t%d" % self.tmp

    @staticmethod
    def _from_c(c_expr):
        return RdTmp(c_expr.Iex.RdTmp.tmp)

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_RdTmp(expr.tmp)

    def result_type(self, tyenv):
        return tyenv.lookup(self.tmp)


class Get(IRExpr):
    """
    Read a guest register, at a fixed offset in the guest state.
    """

    __slots__ = ['offset', 'ty']

    tag = 'Iex_Get'

    def __init__(self, offset, ty):
        IRExpr.__init__(self)
        self.offset = offset
        self.ty = ty

    @property
    def type(self):
        return self.ty

    def __str__(self, reg_name=None):
        if reg_name:
            return "GET:%s(%s)" % (self.ty[4:], reg_name)
        else:
            return "GET:%s(offset=%s)" % (self.ty[4:], self.offset)

    @staticmethod
    def _from_c(c_expr):
        return Get(c_expr.Iex.Get.offset,
                   get_enum_from_int(c_expr.Iex.Get.ty))

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Get(expr.offset,
                              get_int_from_enum(expr.ty))

    def result_type(self, tyenv):
        return self.ty


class Qop(IRExpr):
    """
    A quaternary operation (4 arguments).
    """

    __slots__ = ['op', 'args']

    tag = 'Iex_Qop'

    def __init__(self, op, args):
        IRExpr.__init__(self)
        self.op = op
        self.args = args

    def __str__(self):
        return "%s(%s)" % (self.op[4:], ','.join(str(a) for a in self.args))

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [ ])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        return Qop(get_enum_from_int(c_expr.Iex.Qop.details.op),
                   [IRExpr._from_c(arg)
                    for arg in [c_expr.Iex.Qop.details.arg1,
                                c_expr.Iex.Qop.details.arg2,
                                c_expr.Iex.Qop.details.arg3,
                                c_expr.Iex.Qop.details.arg4]])

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Qop(get_int_from_enum(expr.op),
                              *[IRExpr._to_c(arg)
                                for arg in expr.args])

    def result_type(self, tyenv):
        return get_op_retty(self.op)

    def typecheck(self, tyenv): # TODO change all this to use PyvexTypeErrorException
        resty, (arg1ty, arg2ty, arg3ty, arg4ty) = op_arg_types(self.op)
        arg1ty_real = self.args[0].typecheck(tyenv)
        arg2ty_real = self.args[1].typecheck(tyenv)
        arg3ty_real = self.args[2].typecheck(tyenv)
        arg4ty_real = self.args[3].typecheck(tyenv)
        if arg1ty_real is None or arg2ty_real is None or arg3ty_real is None or arg4ty_real is None:
            return None

        if arg1ty_real != arg1ty:
            l.debug("First arg of %s must be %s", self.op, arg1ty)
            return None
        if arg2ty_real != arg2ty:
            l.debug("Second arg of %s must be %s", self.op, arg2ty)
            return None
        if arg3ty_real != arg3ty:
            l.debug("Third arg of %s must be %s", self.op, arg3ty)
            return None
        if arg4ty_real != arg4ty:
            l.debug("Fourth arg of %s must be %s", self.op, arg4ty)
            return None

        return resty

class Triop(IRExpr):
    """
    A ternary operation (3 arguments)
    """

    __slots__ = ['op', 'args']

    tag = 'Iex_Triop'

    def __init__(self, op, args):
        IRExpr.__init__(self)
        self.op = op
        self.args = args

    def __str__(self):
        return "%s(%s)" % (self.op[4:], ','.join(str(a) for a in self.args))

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [ ])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        return Triop(get_enum_from_int(c_expr.Iex.Triop.details.op),
                     [IRExpr._from_c(arg)
                      for arg in [c_expr.Iex.Triop.details.arg1,
                                  c_expr.Iex.Triop.details.arg2,
                                  c_expr.Iex.Triop.details.arg3]])

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Triop(get_int_from_enum(expr.op),
                                *[IRExpr._to_c(arg)
                                  for arg in expr.args])

    def result_type(self, tyenv):
        return get_op_retty(self.op)

    def typecheck(self, tyenv):
        resty, (arg1ty, arg2ty, arg3ty) = op_arg_types(self.op)
        arg1ty_real = self.args[0].typecheck(tyenv)
        arg2ty_real = self.args[1].typecheck(tyenv)
        arg3ty_real = self.args[2].typecheck(tyenv)
        if arg1ty_real is None or arg2ty_real is None or arg3ty_real is None:
            return None

        if arg1ty_real != arg1ty:
            l.debug("First arg of %s must be %s", self.op, arg1ty)
            return None
        if arg2ty_real != arg2ty:
            l.debug("Second arg of %s must be %s", self.op, arg2ty)
            return None
        if arg3ty_real != arg3ty:
            l.debug("Third arg of %s must be %s", self.op, arg3ty)
            return None

        return resty


class Binop(IRExpr):
    """
    A binary operation (2 arguments).
    """

    __slots__ = ['op', 'args']

    tag = 'Iex_Binop'

    def __init__(self, op, args):
        IRExpr.__init__(self)
        self.op = op
        self.args = args

    def __str__(self):
        return "%s(%s)" % (self.op[4:], ','.join(str(a) for a in self.args))

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [ ])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        return Binop(get_enum_from_int(c_expr.Iex.Binop.op),
                     [IRExpr._from_c(arg)
                      for arg in [c_expr.Iex.Binop.arg1,
                                  c_expr.Iex.Binop.arg2]])

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Binop(get_int_from_enum(expr.op),
                                *[IRExpr._to_c(arg)
                                  for arg in expr.args])

    def result_type(self, tyenv):
        return get_op_retty(self.op)

    def typecheck(self, tyenv):
        arg1ty_real = self.args[0].typecheck(tyenv)
        arg2ty_real = self.args[1].typecheck(tyenv)

        retty = get_op_retty(self.op, arg1ty_real, arg2ty_real)

        resty, (arg1ty, arg2ty) = op_arg_types(self.op)
        if arg1ty_real is None or arg2ty_real is None:
            return None

        if arg1ty_real != arg1ty:
            l.debug("First arg of %s must be %s", self.op, arg1ty)
            return None
        if arg2ty_real != arg2ty:
            l.debug("Second arg of %s must be %s", self.op, arg2ty)
            return None

        return resty


class Unop(IRExpr):
    """
    A unary operation (1 argument).
    """

    __slots__ = ['op', 'args']

    tag = 'Iex_Unop'

    def __init__(self, op, args):
        IRExpr.__init__(self)
        self.op = op
        self.args = args

    def __str__(self):
        return "%s(%s)" % (self.op[4:], ','.join(str(a) for a in self.args))

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [ ])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        return Unop(get_enum_from_int(c_expr.Iex.Unop.op),
                    [IRExpr._from_c(c_expr.Iex.Unop.arg)])

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Unop(get_int_from_enum(expr.op),
                               IRExpr._to_c(expr.args[0]))

    def result_type(self, tyenv):
        return get_op_retty(self.op)

    def typecheck(self, tyenv):
        resty, (arg1ty,) = op_arg_types(self.op)
        arg1ty_real = self.args[0].typecheck(tyenv)
        if arg1ty_real is None:
            return None

        if arg1ty_real != arg1ty:
            l.debug("First arg of %s must be %s", self.op, arg1ty)
            return None

        return resty


class Load(IRExpr):
    """
    A load from memory.
    """

    __slots__ = ['end', 'ty', 'addr']

    tag = 'Iex_Load'

    def __init__(self, end, ty, addr):
        IRExpr.__init__(self)
        self.end = end
        self.ty = ty
        self.addr = addr

    @property
    def endness(self):
        return self.end

    @property
    def type(self):
        return self.ty

    def __str__(self):
        return "LD%s:%s(%s)" % (self.end[-2:].lower(), self.ty[4:], self.addr)

    @staticmethod
    def _from_c(c_expr):
        return Load(get_enum_from_int(c_expr.Iex.Load.end),
                    get_enum_from_int(c_expr.Iex.Load.ty),
                    IRExpr._from_c(c_expr.Iex.Load.addr))

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Load(get_int_from_enum(expr.end),
                               get_int_from_enum(expr.ty),
                               IRExpr._to_c(expr.addr))

    def result_type(self, tyenv):
        return self.ty

    def typecheck(self, tyenv):
        addrty = self.addr.typecheck(tyenv)
        if addrty is None:
            return None
        if addrty != tyenv.wordty:
            l.debug("Address must be word-sized")
            return None
        return self.ty


class Const(IRExpr):
    """
    A constant expression.
    """

    __slots__ = ['con']

    tag = 'Iex_Const'

    def __init__(self, con):
        IRExpr.__init__(self)
        self.con = con

    def __str__(self):
        return str(self.con)

    @staticmethod
    def _from_c(c_expr):
        return Const(IRConst._from_c(c_expr.Iex.Const.con))

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Const(IRConst._to_c(expr.con))

    def result_type(self, tyenv):
        return self.con.type


class ITE(IRExpr):
    """
    An if-then-else expression.
    """

    __slots__ = ['cond', 'iffalse', 'iftrue']

    tag = 'Iex_ITE'

    def __init__(self, cond, iffalse, iftrue):
        IRExpr.__init__(self)
        self.cond = cond
        self.iffalse = iffalse
        self.iftrue = iftrue

    def __str__(self):
        return "ITE(%s,%s,%s)" % (self.cond, self.iftrue, self.iffalse)

    @staticmethod
    def _from_c(c_expr):
        return ITE(IRExpr._from_c(c_expr.Iex.ITE.cond),
                   IRExpr._from_c(c_expr.Iex.ITE.iffalse),
                   IRExpr._from_c(c_expr.Iex.ITE.iftrue))

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_ITE(IRExpr._to_c(expr.cond),
                              IRExpr._to_c(expr.iftrue),
                              IRExpr._to_c(expr.iffalse))

    def result_type(self, tyenv):
        return self.iftrue.result_type(tyenv)

    def typecheck(self, tyenv):
        condty = self.cond.typecheck(tyenv)
        falsety = self.iffalse.typecheck(tyenv)
        truety = self.iftrue.typecheck(tyenv)

        if condty is None or falsety is None or truety is None:
            return None

        if condty != 'Ity_I1':
            l.debug("guard must be Ity_I1")
            return None

        if falsety != truety:
            l.debug("false condition must be same type as true condition")
            return None

        return falsety

class CCall(IRExpr):
    """
    A call to a pure (no side-effects) helper C function.
    """

    __slots__ = ['retty', 'cee', 'args']

    tag = 'Iex_CCall'

    def __init__(self, retty, cee, args):
        IRExpr.__init__(self)
        self.retty = retty
        self.cee = cee
        self.args = tuple(args)

    @property
    def ret_type(self):
        return self.retty

    @property
    def callee(self):
        return self.cee

    def __str__(self):
        return "%s(%s):%s" % (self.cee, ','.join(str(a) for a in self.args), self.retty)

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [ ])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        i = 0
        args = []
        while True:
            arg = c_expr.Iex.CCall.args[i]
            if arg == ffi.NULL:
                break
            args.append(IRExpr._from_c(arg))
            i += 1

        return CCall(get_enum_from_int(c_expr.Iex.CCall.retty),
                     IRCallee._from_c(c_expr.Iex.CCall.cee),
                     tuple(args))

    @staticmethod
    def _to_c(expr):
        args = [IRExpr._to_c(arg) for arg in expr.args]
        mkIRExprVec = getattr(pvc, 'mkIRExprVec_%d' % len(args))
        return pvc.IRExpr_CCall(IRCallee._to_c(expr.cee),
                                get_int_from_enum(expr.retty),
                                mkIRExprVec(*args))

    def result_type(self, tyenv):
        return self.retty

def get_op_retty(op, *args):
    return op_arg_types(op)[0]

op_signatures = {}
def _request_op_type_from_cache(op):
    return op_signatures[op]

def _request_op_type_from_libvex(op):
    res_ty = ffi.new('IRType *')
    arg_tys = [ffi.new('IRType *') for _ in xrange(4)]
    for arg in arg_tys:
        arg = 0x1100
    pvc.typeOfPrimop(get_int_from_enum(op), res_ty, *arg_tys)
    arg_ty_vals = [a[0] for a in arg_tys]

    try:
        numargs = arg_ty_vals.index(0x1100)
    except ValueError:
        numargs = 4
    args_tys_list = [get_enum_from_int(arg_ty_vals[i]) for i in range(numargs)]

    op_ty_sig = (get_enum_from_int(res_ty[0]), tuple(args_tys_list))
    op_signatures[op] = op_ty_sig
    return op_ty_sig

class PyvexOpMatchException(Exception):
    pass

class PyvexTypeErrorException(Exception):
    pass

def int_type_for_size(size):
    return 'Ity_I%d' % size

def unop_signature(op):
    m = re.match(r'Iop_(Not|Ctz|Clz)(?P<size>\d+)$', op)
    if m is None:
        raise PyvexOpMatchException()
    size = int(m.group('size'))
    size_type = int_type_for_size(size)
    return (size_type, (size_type,))

def binop_signature(op):
    m = re.match(r'Iop_(Add|Sub|Mul|Xor|Or|And|Div(S|U))(?P<size>\d+)$', op)
    if m is None:
        raise PyvexOpMatchException()
    size = int(m.group('size'))
    size_type = int_type_for_size(size)
    return (size_type, (size_type, size_type))

def shift_signature(op):
    m = re.match(r'Iop_(Shl|Shr|Sar)(?P<size>\d+)$', op)
    if m is None:
        raise PyvexOpMatchException()
    size = int(m.group('size'))
    if size > 255:
        raise PyvexTypeErrorException('Cannot apply shift operation to %d size int because shift index is 8-bit' % size)
    size_type = int_type_for_size(size)
    return (size_type, (size_type, int_type_for_size(8)))

def cmp_signature(op):
    m = re.match(r'Iop_Cmp(EQ|NE)(?P<size>\d+)$', op)
    m2 = re.match(r'Iop_Cmp(GT|GE|LT|LE)(?P<size>\d+)[SU]$', op)
    if (m is None) == (m2 is None):
        raise PyvexOpMatchException()
    mfound = m if m is not None else m2
    size = int(mfound.group('size'))
    size_type = int_type_for_size(size)
    return (int_type_for_size(1), (size_type, size_type))

def mull_signature(op):
    m = re.match(r'Iop_Mull[SU](?P<size>\d+)$', op)
    if m is None:
        raise PyvexOpMatchException()
    size = int(m.group('size'))
    size_type = int_type_for_size(size)
    doubled_size_type = int_type_for_size(2 * size)
    return (doubled_size_type, (size_type, size_type))

def half_signature(op):
    m = re.match(r'Iop_DivMod[SU](?P<fullsize>\d+)to(?P<halfsize>\d+)$', op)
    if m is None:
        raise PyvexOpMatchException()
    fullsize = int(m.group('fullsize'))
    halfsize = int(m.group('halfsize'))
    if halfsize * 2 != fullsize:
        raise PyvexTypeErrorException('Invalid Instruction %s: Type 1 must be twice the size of type 2' % op)
    fullsize_type = int_type_for_size(fullsize)
    halfsize_type = int_type_for_size(halfsize)
    return (fullsize_type, (fullsize_type, halfsize_type))

def cast_signature(op):
    m = re.match(r'Iop_(?P<srcsize>\d+)(U|S|HI|HL)?to(?P<dstsize>\d+)', op)
    if m is None:
        raise PyvexOpMatchException()
    src_type = int_type_for_size(int(m.group('srcsize')))
    dst_type = int_type_for_size(int(m.group('dstsize')))
    return (dst_type, (src_type,))

polymorphic_op_processors = [ unop_signature,
                              binop_signature,
                              shift_signature,
                              cmp_signature,
                              mull_signature,
                              half_signature,
                              cast_signature ]

def _request_polymorphic_op_type(op):
    for polymorphic_signature in polymorphic_op_processors:
        try:
            op_ty_sig = polymorphic_signature(op)
            break
        except PyvexOpMatchException:
            continue
    else:
        raise PyvexOpMatchException('Op %s not recognized' % op)
    return op_ty_sig

_request_funcs = [_request_op_type_from_cache,
                  _request_op_type_from_libvex,
                  _request_polymorphic_op_type]

def op_arg_types(op):
    for _request_func in _request_funcs:
        try:
           return _request_func(op)
        except KeyError:
            continue
    raise ValueError("Cannot find type of op %s" % op)

from .const import IRConst
from .errors import PyVEXError
from . import ffi, pvc

def tag_to_expr_class(tag):
    m = re.match(r'Iex_(?P<classname>.*)', tag)
    try:
        return globals()[m.group('classname')]
    except TypeError, KeyError:
        raise ValueError('Cannot find expression class for type %s' % tag)



lifters = []
postprocessors = []


class Lifter(object):
    """
    A lifter is a class of methods for processing a block.
    """
    REQUIRE_DATA_C = False
    REQUIRE_DATA_PY = False

    def __init__(self,
            irsb,
            data,
            max_inst,
            max_bytes,
            bytes_offset,
            opt_level,
            traceflags,
            allow_lookback):
        self.irsb = irsb
        self.data = data
        self.max_inst = max_inst
        self.max_bytes = max_bytes
        self.bytes_offset = bytes_offset
        self.opt_level = opt_level
        self.traceflags = traceflags
        self.allow_lookback = allow_lookback

        self._error = None  # you might want to use this

    # pylint: disable=no-self-use

    def lift(self):
        """
        Populate the fields of the empty IRSB passed in. Return whether or not successful.

        Only the first successful lifter will be used.
        """
        return False

    def error(self):
        """
        Return the error message from an unsuccessful lift. If the error is just "unsupported
        architecture", just return None.
        """
        return self._error

    def postprocess(self):
        """
        Postprocess the populated IRSB passed in.
        """
        pass

# max_bytes should always be provided
# max_inst might be none
def lift(irsb, data, max_bytes, max_inst, bytes_offset, opt_level, traceflags):
    if not max_bytes and not isinstance(data, (str, bytes)):
        raise PyVEXError("C-backed bytes must have the length specified by max_bytes")
    if not max_bytes:
        max_bytes = len(data)

    if max_bytes == 0:
        raise PyVEXError("No bytes provided")

    errors = []

    if isinstance(data, (str, bytes)):
        py_data = data
        c_data = None
        allow_lookback = False
    else:
        c_data = data
        py_data = None
        allow_lookback = True

    for lifter in lifters:
        u_data = data
        if lifter.REQUIRE_DATA_C:
            if c_data is None:
                c_data = ffi.new('unsigned char [%d]' % (len(data) + 8), data + '\0' * 8)
            u_data = c_data
        elif lifter.REQUIRE_DATA_PY:
            if py_data is None:
                py_data = str(ffi.buffer(data, max_bytes))
            u_data = py_data

        # Setting it up with an instance per lift like this allows us thread safety, hypothetically
        # this could be done more efficiently with thread-local vars
        lifter_inst = lifter(irsb, u_data, max_inst, max_bytes, bytes_offset, opt_level, traceflags, allow_lookback)
        if lifter_inst.lift():
            break
        else:
            err = lifter_inst.error()
            if err is not None:
                errors.append(err)
    else:
        if len(errors) == 0:
            raise PyVEXError('No lifters for architecture %s!' % irsb.arch.name)
        raise PyVEXError('\n\n'.join(errors))

    for lifter in postprocessors:
        u_data = data
        if lifter.REQUIRE_DATA_C:
            if c_data is None:
                c_data = ffi.new('unsigned char [%d]' % (len(data) + 8), data + '\0' * 8)
            u_data = c_data
        elif lifter.REQUIRE_DATA_PY:
            if py_data is None:
                py_data = str(ffi.buffer(data, max_bytes))
            u_data = py_data
        lifter_inst = lifter(irsb, u_data, max_inst, max_bytes, bytes_offset, opt_level, traceflags, allow_lookback)
        lifter_inst.postprocess()

def register(lifter):
    if lifter.lift is not Lifter.lift:
        lifters.append(lifter)
    if lifter.postprocess is not Lifter.postprocess:
        postprocessors.append(lifter)

from .. import ffi
from ..errors import PyVEXError

from .libvex import LibVEXLifter
from .fixes import FixesPostProcessor



from . import VEXObject
from . import expr, stmt, const
from .lift import lift
from .enums import get_enum_from_int, get_int_from_enum
from .const import get_type_size
from .errors import PyVEXError

import logging
l = logging.getLogger("pyvex.block")

class IRSB(VEXObject):
    """
    The IRSB is the primary interface to pyvex. Constructing one of these will make a call into LibVEX to perform a
    translation.

    IRSB stands for *Intermediate Representation Super-Block*. An IRSB in VEX is a single-entry, multiple-exit code
    block.

    :ivar arch:             The architecture this block is lifted under
    :vartype arch:          :class:`archinfo.Arch`
    :ivar statements:       The statements in this block
    :vartype statements:    list of :class:`IRStmt`
    :ivar next:             The expression for the default exit target of this block
    :vartype next:          :class:`IRExpr`
    :ivar int offsIP:       The offset of the instruction pointer in the VEX guest state
    :ivar int stmts_used:   The number of statements in this IRSB
    :ivar str jumpkind:     The type of this block's default jump (call, boring, syscall, etc) as a VEX enum string
    :ivar bool direct_next: Whether this block ends with a direct (not indirect) jump or branch
    :ivar int size:         The size of this block in bytes
    :ivar int addr:         The address of this basic block, i.e. the address in the first IMark
    """

    __slots__ = ['_addr', 'arch', 'statements', 'next', 'tyenv', 'jumpkind', '_direct_next', '_size']

    def __init__(self, data, mem_addr, arch, max_inst=None, max_bytes=None, bytes_offset=0, traceflags=0, opt_level=1, num_inst=None, num_bytes=None):
        """
        :param data:            The bytes to lift. Can be either a string of bytes or a cffi buffer object.
                                You may also pass None to initialize an empty IRSB.
        :type data:             str or bytes or cffi.FFI.CData or None
        :param int mem_addr:    The address to lift the data at.
        :param arch:            The architecture to lift the data as.
        :type arch:             :class:`archinfo.Arch`
        :param max_inst:        The maximum number of instructions to lift. Max 99. (See note below)
        :param max_bytes:       The maximum number of bytes to use. Max 5000.
        :param bytes_offset:    The offset into `data` to start lifting at.
        :param traceflags:      The libVEX traceflags, controlling VEX debug prints.
        :param opt_level:       The level of optimization to apply to the IR, 0-2.

        .. note:: Explicitly specifying the number of instructions to lift (`max_inst`) may not always work
                  exactly as expected. For example, on MIPS, it is meaningless to lift a branch or jump
                  instruction without its delay slot. VEX attempts to Do The Right Thing by possibly decoding
                  fewer instructions than requested. Specifically, this means that lifting a branch or jump
                  on MIPS as a single instruction (`max_inst=1`) will result in an empty IRSB, and subsequent
                  attempts to run this block will raise `SimIRSBError('Empty IRSB passed to SimIRSB.')`.
        """
        if max_inst is None: max_inst = num_inst
        if max_bytes is None: max_bytes = num_bytes
        VEXObject.__init__(self)
        self._addr = mem_addr
        self.arch = arch

        self.statements = []
        self.next = None
        self.tyenv = IRTypeEnv(arch)
        self.jumpkind = None
        self._direct_next = None
        self._size = None

        if data is not None:
            lift(self, data, max_bytes, max_inst, bytes_offset, opt_level, traceflags)

    def pp(self):
        """
        Pretty-print the IRSB to stdout.
        """
        print self._pp_str()

    def typecheck(self):
        try:
            # existence assertions
            assert self.next is not None, "Missing next expression"
            assert self.jumpkind is not None, "Missing jumpkind"

            # type assertions
            assert isinstance(self.next, expr.IRExpr), "Next expression is not an expression"
            assert type(self.jumpkind is str), "Jumpkind is not a string"
            assert self.jumpkind.startswith('Ijk_'), "Jumpkind is not a jumpkind enum"
            assert self.tyenv.typecheck(), "Type environment contains invalid types"

            # statement assertions
            last_imark = None
            for i, st in enumerate(self.statements):
                assert isinstance(st, stmt.IRStmt), "Statement %d is not an IRStmt" % i
                try:
                    assert st.typecheck(self.tyenv), "Statement %d failed to typecheck" % i
                except: # pylint: disable=bare-except
                    assert False, "Statement %d errored in typechecking" % i

                if type(st) is stmt.NoOp:
                    continue
                elif type(st) is stmt.IMark:
                    if last_imark is not None:
                        # pylint: disable=unsubscriptable-object
                        assert last_imark[0] + last_imark[1] == st.addr, "IMarks sizes overlap or have gaps"
                    last_imark = (st.addr, st.len)
                else:
                    assert last_imark is not None, "Operation statement appears before IMark"

            assert last_imark is not None, "No IMarks present in block"
        except AssertionError as e:
            l.debug(e.args[0])
            return False
        return True

    #
    # alternate constructors
    #

    @staticmethod
    def from_c(c_irsb, mem_addr, arch):
        irsb = IRSB(None, mem_addr, arch)
        irsb._from_c(c_irsb)
        return irsb

    @staticmethod
    def from_py(tyenv, stmts, next_expr, jumpkind, mem_addr, arch):
        irsb = IRSB(None, mem_addr, arch)

        irsb.tyenv = tyenv
        irsb.statements = stmts
        irsb.next = next_expr
        irsb.jumpkind = jumpkind
        irsb._direct_next = irsb._is_defaultexit_direct_jump()

        return irsb

    #
    # simple properties useful for analysis
    #

    @property
    def stmts_used(self):
        return len(self.statements)

    @property
    def offsIP(self):
        return self.arch.ip_offset

    @property
    def direct_next(self):
        if self._direct_next is None:
            self._direct_next = self._is_defaultexit_direct_jump()
        return self._direct_next

    @property
    def expressions(self):
        """
        A list of all expressions contained in the IRSB.
        """
        expressions = []
        for s in self.statements:
            expressions.extend(s.expressions)
        expressions.append(self.next)
        return expressions

    @property
    def instructions(self):
        """
        The number of instructions in this block
        """
        return len([s for s in self.statements if isinstance(s, stmt.IMark)])

    @property
    def size(self):
        """
        The size of this block, in bytes
        """
        if self._size is None:
            self._size = sum([s.len for s in self.statements if isinstance(s, stmt.IMark)])
        return self._size

    @property
    def addr(self):
        for s in self.statements:
            if isinstance(s, stmt.IMark):
                return s.addr
        return None

    @property
    def operations(self):
        """
        A list of all operations done by the IRSB, as libVEX enum names
        """
        ops = []
        for e in self.expressions:
            if hasattr(e, 'op'):
                ops.append(e.op)
        return ops

    @property
    def all_constants(self):
        """
        Returns all constants in the block (including incrementing of the program counter) as :class:`pyvex.const.IRConst`.
        """
        return sum((e.constants for e in self.expressions), [])

    @property
    def constants(self):
        """
        The constants (excluding updates of the program counter) in the IRSB as :class:`pyvex.const.IRConst`.
        """
        return sum(
            (s.constants for s in self.statements if not (isinstance(s, stmt.Put) and s.offset == self.offsIP)), [])

    @property
    def constant_jump_targets(self):
        """
        A set of the static jump targets of the basic block.
        """
        exits = set()
        for s in self.statements:
            if isinstance(s, stmt.Exit):
                exits.add(s.dst.value)

        default_target = self._get_defaultexit_target()
        if default_target is not None:
            exits.add(default_target)

        return exits

    @property
    def constant_jump_targets_and_jumpkinds(self):
        """
        A dict of the static jump targets of the basic block to their jumpkind.
        """
        exits = dict()
        for s in self.statements:
            if isinstance(s, stmt.Exit):
                exits[s.dst.value] = s.jumpkind

        default_target = self._get_defaultexit_target()
        if default_target is not None:
            exits[default_target] = self.jumpkind

        return exits

    #
    # private methods
    #

    def _pp_str(self):
        """
        Return the pretty-printed IRSB.

        :rtype: str
        """
        sa = []
        sa.append("IRSB {")
        sa.append("   %s" % self.tyenv)
        sa.append("")
        for i, s in enumerate(self.statements):
            stmt_str = ''
            if isinstance(s, stmt.Put):
                stmt_str = s.__str__(reg_name=self.arch.translate_register_name(s.offset, s.data.result_size(self.tyenv)/8))
            elif isinstance(s, stmt.WrTmp) and isinstance(s.data, expr.Get):
                stmt_str = s.__str__(reg_name=self.arch.translate_register_name(s.data.offset, s.data.result_size(self.tyenv)/8))
            elif isinstance(s, stmt.Exit):
                stmt_str = s.__str__(reg_name=self.arch.translate_register_name(s.offsIP, self.arch.bits / 8))
            else:
                stmt_str = s.__str__()
            sa.append("   %02d | %s" % (i, stmt_str))
        sa.append(
            "   NEXT: PUT(%s) = %s; %s" % (self.arch.translate_register_name(self.offsIP), self.next, self.jumpkind))
        sa.append("}")
        return '\n'.join(sa)

    def _get_defaultexit_target(self):
        """
        The default exit target, if it is constant, or None.
        """
        if isinstance(self.next, expr.Const):
            return self.next.con.value

        if not isinstance(self.next, expr.RdTmp):
            raise PyVEXError("unexpected self.next type: %s", self.next.__class__.__name__)

        tmp_next = self.next.tmp
        reg_next = None
        reg_next_size = None
        for stat in reversed(self.statements):
            if isinstance(stat, stmt.WrTmp) and stat.tmp == tmp_next:
                data = stat.data
            elif isinstance(stat, stmt.Put) and stat.offset == reg_next:
                data = stat.data
                if data.result_size(self.tyenv) != reg_next_size:
                    return None
            elif isinstance(stat, stmt.LoadG) and stat.dst == tmp_next:
                return None
            else:
                continue

            if isinstance(data, expr.Const):
                return data.con.value
            elif isinstance(data, expr.RdTmp):
                tmp_next = data.tmp
                reg_next = None
            elif isinstance(data, expr.Get):
                tmp_next = None
                reg_next = data.offset
                reg_next_size = data.result_size(self.tyenv)
            else:
                return None

        if tmp_next is not None:
            raise PyVEXError('Malformed IRSB at address #%x. Please report to Fish.' % self._addr)
        return None

    def _is_defaultexit_direct_jump(self):
        """
        Checks if the default of this IRSB a direct jump or not.
        """
        if not (self.jumpkind == 'Ijk_Boring' or self.jumpkind == 'Ijk_Call'):
            return False

        target = self._get_defaultexit_target()
        return target is not None

    #
    # internal "constructors" to fill this block out with data from various sources
    #

    def _from_c(self, c_irsb):
        self.statements = [stmt.IRStmt._from_c(c_irsb.stmts[i]) for i in xrange(c_irsb.stmts_used)]
        self.tyenv = IRTypeEnv._from_c(self.arch, c_irsb.tyenv)
        self.next = expr.IRExpr._from_c(c_irsb.next)
        self.jumpkind = get_enum_from_int(c_irsb.jumpkind)

class IRTypeEnv(VEXObject):
    """
    An IR type environment.

    :ivar types:        A list of the types of all the temporaries in this block as VEX enum strings.
                        `types[3]` is the type of t3.
    :vartype types:     list of str
    """

    __slots__ = ['types', 'wordty']

    def __init__(self, arch, types=None):
        VEXObject.__init__(self)
        self.types = [] if types is None else types
        self.wordty = 'Ity_I%d' % arch.bits

    def __str__(self):
        return ' '.join(("t%d:%s" % (i, t)) for i, t in enumerate(self.types))

    def lookup(self, tmp):
        """
        Return the type of temporary variable `tmp` as an enum string
        """
        if tmp < 0 or tmp > self.types_used:
            l.debug("Invalid temporary number %d", tmp)
            raise IndexError(tmp)
        return self.types[tmp]

    def sizeof(self, tmp):
        return get_type_size(self.lookup(tmp))

    def add(self, ty):
        """
        Add a new tmp of type `ty` to the environment. Returns the number of the new tmp.
        """
        self.types.append(ty)
        return self.types_used - 1

    @property
    def types_used(self):
        return len(self.types)

    @staticmethod
    def _from_c(arch, c_tyenv):
        return IRTypeEnv(arch, [get_enum_from_int(c_tyenv.types[t]) for t in xrange(c_tyenv.types_used)])

    @staticmethod
    def _to_c(tyenv):
        c_tyenv = pvc.emptyIRTypeEnv()
        for ty in tyenv.types:
            pvc.newIRTemp(c_tyenv, get_int_from_enum(ty))
        return c_tyenv

    def typecheck(self):
        for ty in self.types:
            try:
               get_type_size(ty)
            except ValueError:
                return False



from lifter_helper import ParseError
from syntax_wrapper import VexValue
from vex_helper import JumpKind, IRExpr

import abc
import string
import bitstring
import logging
from angr.engines.vex import ccall

l = logging.getLogger("instr")
l.setLevel(logging.DEBUG)


class Instruction:
    """
    Base class for an Instruction.
    You should make a subclass of this for each instruction you want to lift.
    These classes will contain the "semantics" of the instruction, that is, what it _does_, in terms of the VEX IR.

    You may want to subclass this for your architecture, and add arch-specific handling
    for parsing, argument resolution, etc, and have instructions subclass that instead.

    The core parsing functionality is done via a "bit format".  Each instruction should be a subclass of Instruction,
    and will be parsed by comparing bits in the provided bitstream to symbols in the bit_format member of the class.
    Bit formats are strings of symbols, like those you'd find in an ISA document, such as "0010rrrrddddffmm"
    0 or 1 specify hard-coded bits that must match for an instruction to match.
    Any letters specify arguments, grouped by letter, which will be parsed and provided as bitstrings in the "data"
    member of the class as a dictionary.
    So, in our example, the bits 0010110101101001, applied to format string 0010rrrrddddffmm
    will result in the following in self.data:
    {'r': '1101',
     'd': '0110',
     'f': '10',
     'm': '01'}

    Implement compute_result to provide the "meat" of what your instruction does.
     You can also implement it in your arch-specific subclass of Instruction, to handle things common to all
     instructions, and provide instruction implementations elsewhere..

    We provide the VexValue syntax wrapper to make expressing instruction semantics easy.
    You first convert the bitstring arguments into VexValues using the provided convenience methods (self.get/put/load)
    store/etc. This loads the register from the actual registers into a temporary value we can work with.
    You can then write it back to a register when you're done.
    For example, if you have the register in 'r', as above, you can make a VexValue like this:
    r_vv = self.get(int(self.data['r'], 2), Type.int_32)
    If you then had an instruction to increment r, you could simply:
    return r_vv += 1
    You could then write it back to the register like this:
    self.put(r_vv, int(self.data['r', 2))

    Note that most architectures have special flags that get set differently for each instruction, make sure to
    implement those as well. (override set_flags() )

    Override parse() to extend parsing; for example, in MSP430, this allows us to grab extra words from the bitstream
    when extra immediate words are present.

    All architectures are different enough that there's no magic recipe for how to write a lifter;
    See the examples provided by gymrat for ideas of how to use this to build your own lifters quickly and easily.
    """

    data = None
    irsb_c = None

    def __init__(self, irsb_c, bitstrm, addr):
        """
        Create an instance of the instruction
        :param irsb_c: The IRSBCustomizer to put VEX instructions into
        :param bitstrm: The bitstream to decode instructions from
        :param addr: The address of the instruction to be lifted, used only for jumps and branches
        """
        self.irsb_c = irsb_c
        self.addr = addr
        self.width = len(self.bin_format)
        self.data = self.parse(bitstrm)

    def __call__(self):
        self.lift()

    def mark_instruction_start(self):
        # TODO: WARNING: VEX assumes 8-bit bytes here.
        bytewidth = self.bitwidth / 8
        self.irsb_c.imark(self.addr, bytewidth, bytewidth)

    def fetch_operands(self):
        """
        Get the operands out of memory or registers
        Return a tuple of operands for the instruction
        :return:
        """
        return []

    def lift(self):
        """
        THis is the main body of the "lifting" for the instruction.
        This can/should be overriden to provide the general flow of how instructions in your arch work.
        For example, in MSP430, this is:
            1) Figure out what your operands are by parsing the addressing, and load them into temporary registers
            2) Do the actual operation, and commit the result, if needed.
            3) Compute the flags
        :return:
        """
        # Always call this first!
        self.mark_instruction_start()
        # Then do the actual stuff.
        inputs = self.fetch_operands()
        retval = self.compute_result(*inputs)
        if retval is not None:
            inputs = tuple(list(inputs).append(retval))
        self.compute_flags(*inputs)

    @abc.abstractmethod
    def compute_result(self, *args):
        """
        This is where the actual operation performed by your instruction, excluding the calculation of flags, should be
        performed.  Return the VexValue of the "result" of the instruction, which may
        be used to calculate the flags later.
        For example, for a simple add, with arguments src and dst, you can simply write:
            return src + dst:

        :param args:
        :return: A VexValue containing the "result" of the operation.
        """
        pass

    def compute_flags(self, *args):
        """
        Most CPU architectures have "flags" that should be computed for many instructions.
        Override this to specify how that happens.  One common pattern is to define this method to call specifi methods
        to update each flag, which can then be overriden in the actual classes for each instruction.
        :return: n/a
        """
        pass

    def match_instruction(self, data, bitstrm):
        """
        Override this to extend the parsing functionality.
        This is great for if your arch has instruction "formats" that have an opcode that has to match.
        :param data:
        :param bitstrm:
        :return: data
        """
        return data

    def parse(self, bitstrm):
        beforebits = bitstrm.bitpos
        numbits = len(self.bin_format)
        if self.irsb_c.irsb.arch.memory_endness == 'Iend_LE':
            # Get it out little endian.  I hate this.
            instr_bits = bitstring.Bits(uint=bitstrm.peek("uintle:%d" % numbits), length=numbits).bin
        else:
            instr_bits = bitstrm.peek("bin:%d" % numbits)
        data = {c : '' for c in self.bin_format if c in string.ascii_letters}
        for c, b in zip(self.bin_format, instr_bits):
            if c in '01':
                if b != c:
                    raise ParseError('Mismatch between format bit %c and instruction bit %c' % (c, b))
            elif c in string.ascii_letters:
                data[c] += b
            else:
                raise ValueError('Invalid bin_format character %c' % c)
        # Hook here for extra matching functionality
        if hasattr(self, 'match_instruction'):
            # Should raise if it's not right
            self.match_instruction(data, bitstrm)
        # Use up the bits once we're sure it's right
        self.rawbits = bitstrm.read('hex:%d' % numbits)
        # Hook here for extra parsing functionality (e.g., trailers)
        if hasattr(self, '_extra_parsing'):
            data = self._extra_parsing(data, bitstrm)
        afterbits = bitstrm.bitpos
        self.bitwidth = afterbits - beforebits
        return data

    def disassemble(self):
        """
        Return the disassembly of this instruction, as a string.
        Override this in subclasses.
        :return: The address (self.addr), the instruction's name, and a list of its operands, as strings
        """
        return self.addr, 'UNK', [self.rawbits]

    # These methods should be called in subclasses to do register and memory operations

    def load(self, addr, ty):
        """
        Load a value from memory into a VEX temporary register.
        :param addr: The VexValue containing the addr to load from.
        :param ty: The Type of the resulting data
        :return: a VexValue
        """
        rdt = self.irsb_c.load(addr.rdt, ty)
        return VexValue(self.irsb_c, rdt)

    def constant(self, val, ty):
        """
        Creates a constant as a VexValue
        :param val: The value, as an integer
        :param ty: The type of the resulting VexValue
        :return: a VexValue
        """
        assert not (isinstance(val, VexValue) or isinstance(val, IRExpr))
        rdt = self.irsb_c.mkconst(val, ty)
        return VexValue(self.irsb_c, rdt)

    def get(self, reg_num, ty):
        """
        Load a value from a machine register into a VEX temporary register.
        All values must be loaded out of registers before they can be used with operations, etc
        and stored back into them when the instruction is over.  See Put().

        :param reg_num: Register number as an integer to get from
        :param ty: The Type to use.
        :return: A VexValue of the gotten value.
        """
        # TODO: Resolve strings like 'sr' or 'pc' into the right numbers using archinfo
        rdt = self.irsb_c.rdreg(reg_num, ty)
        return VexValue(self.irsb_c, rdt)

    def put(self, val, reg):
        """
        Puts a value from a VEX temporary register into a machine register.
        This is how the results of operations done to registers get committed to the machine's state.
        :param val: The VexValue to store (Want to store a constant? See Constant() first)
        :param reg: The integer register number to store into.
        :return: None
        """
        # TODO: Resolve strings like 'sr' or 'pc' into the right numbers using archinfo
        self.irsb_c.put(val.rdt, reg)

    def store(self, val, addr):
        """
        Store a VexValue in memory at the specified loaction.
        :param val: The VexValue of the value to store
        :param addr: The VexValue of the address to store into
        :return: None
        """
        # TODO: Resolve strings like 'sr' or 'pc' into the right numbers using archinfo
        self.irsb_c.store(addr.rdt, val.rdt)

    def jump(self, condition, to_addr, jumpkind=JumpKind.Boring):
        """
        Jump to a specified destination, under the specified condition.
        Used for branches, jumps, calls, returns, etc.
        :param condition: The VexValue representing the expression for the guard, or None for an unconditional jump
        :param to_addr: The address to jump to.
        :param jumpkind: The JumpKind to use.  See the VEX docs for what these are; you only need them for things
        aren't normal jumps (e.g., calls, interrupts, program exits, etc etc)
        :return: None
        """
        assert isinstance(to_addr, VexValue)
        if not condition:
            # This is the default exit.
            self.irsb_c.irsb.jumpkind = jumpkind
            self.irsb_c.irsb.next = to_addr.rdt
        else:
            # add another exit
            self.irsb_c.add_exit(condition.rdt, to_addr.rdt.con, jumpkind, self.addr)
            # and then set the default
            self.irsb_c.irsb.jumpkind = jumpkind
            self.irsb_c.irsb.next = self.constant(self.addr + (self.bitwidth / 8), to_addr.ty).rdt

    def ccall(self, ret_type, func_obj, args):
        """
        Creates a CCall operation.
        A CCall is a procedure that calculates a value at *runtime*, not at lift-time.
        You can use these for flags, unresolvable jump targets, etc.
        We caution you to avoid using them when at all possible though.

        For an example of how to write and use a CCall, see gymrat/bf/lift_bf.py
        :param ret_type: The return type of the CCall
        :param func_obj: The function object to eventually call.
        :param args: List of arguments to the function
        :return: A VexValue of the result.
        """
        # HACK: FIXME: If you're reading this, I'm sorry. It's truly a crime against Python...
        if not hasattr(ccall, func_obj.func_name):
            setattr(ccall, func_obj.func_name, func_obj)
        cc = self.irsb_c.op_ccall(ret_type, func_obj.func_name, args)
        return VexValue(self.irsb_c, cc)
